/**
 * This file is part of Princess, a theorem prover for Presburger
 * arithmetic with uninterpreted predicates.
 * <http://www.philipp.ruemmer.org/princess.shtml>
 *
 * Copyright (C) 2009 Philipp Ruemmer <ph_r@gmx.net>
 *
 * Princess is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Princess is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Princess.  If not, see <http://www.gnu.org/licenses/>.
 */

package ap.proof;

import ap.basetypes.IdealInt
import ap.terfor.{Formula, TermOrder, ConstantTerm}
import ap.terfor.conjunctions.{Conjunction, Quantifier, ReduceWithConjunction}
import ap.terfor.linearcombination.LinearCombination
import ap.terfor.equations.EquationConj
import ap.terfor.substitutions.{Substitution, IdentitySubst}
import ap.proof.goal.{Goal, FactsNormalisationTask, EliminateFactsTask,
                      AddFactsTask, UpdateTasksTask}
import ap.proof.certificates.Certificate
import ap.util.{Debug, Logic, LRUCache, FilterIt, Seqs}
import ap.parameters.{GoalSettings, Param}
import ap.proof.tree._

import scala.collection.mutable.Stack


/**
 * A prover that tries to construct a countermodel of a ground formula. This
 * prover works in depth-first mode, in contrast to the
 * <code>ExhaustiveProver</code>.
 */
object ModelSearchProver {

  private val AC = Debug.AC_PROVER
   
  private val simplifier = ConstraintSimplifier.FAIR_SIMPLIFIER
  
  // we need to store eliminated facts from goals, otherwise we could not
  // construct a complete model
  private val ptf = new SimpleProofTreeFactory(true, simplifier) {
    override def eliminatedConstant(subtree : ProofTree,
                                    c : ConstantTerm,
                                    witness : (Substitution, TermOrder) => Substitution,
                                    vocabulary : Vocabulary) : ProofTree =
      new WitnessTree (subtree, c, witness, vocabulary)
  }

  private val cache = new LRUCache[Formula, Conjunction] (1000)
  
  /**
   * <code>inputFor</code> is the formula to be disproven. The result of the
   * method is a countermodel of <code>inputFor</code>, or <code>FALSE</code>
   * if it was not possible to find one (this implies that <code>inputFor</code>
   * is valid).
   */
  def apply(inputFor : Formula, order : TermOrder) : Conjunction = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ModelSearchProver.AC,
                    inputFor.variables.isEmpty && (order isSortingOf inputFor))
    ////////////////////////////////////////////////////////////////////////////
    cache(inputFor) {
      applyHelp(List(Conjunction.conj(inputFor, order)),
                order, GoalSettings.DEFAULT).left.get
    }
  }

  def apply(inputDisjuncts : Seq[Conjunction], order : TermOrder,
            settings : GoalSettings) : Either[Conjunction, Certificate] = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ModelSearchProver.AC,
                    inputDisjuncts forall ((inputFor) =>
                      inputFor.variables.isEmpty && (order isSortingOf inputFor)))
    ////////////////////////////////////////////////////////////////////////////
    applyHelp(inputDisjuncts, order,
              Param.CONSTRAINT_SIMPLIFIER.set(settings, simplifier))
  }
   
  /**
   * <code>inputFor</code> is the formula to be disproven. The result of the
   * method is a countermodel of <code>inputFor</code>, or <code>None</code>
   * if it was not possible to find one (this implies that <code>inputFor</code>
   * is valid).
   */
  private def applyHelp(inputDisjuncts : Seq[Conjunction], order : TermOrder,
                        settings : GoalSettings) : Either[Conjunction, Certificate] = {
    val disjuncts = if (Param.PROOF_CONSTRUCTION(settings)) {
      inputDisjuncts
    } else {
      val reducer = ReduceWithConjunction(Conjunction.TRUE, order)
      for (d <- inputDisjuncts) yield reducer(d)
    }
    val goal = Goal(disjuncts, order.orderedConstants, order, settings)

    //    val model = findModelFair(goal, 500)
    findModel(goal, List(), Set(), 0, settings) match {
      case Left(model) =>
        Left(model)
      case Right(certificates) => {
        if (Param.PROOF_CONSTRUCTION(settings)) {
          //////////////////////////////////////////////////////////////////////
          Debug.assertInt(ModelSearchProver.AC,
                          certificates.size == 1 && certificates(0) != null)
          //////////////////////////////////////////////////////////////////////
          val Seq(cert) = certificates
          //////////////////////////////////////////////////////////////////////
          Debug.assertInt(ModelSearchProver.AC,
                          cert.assumedFormulas subsetOf
                            (Set() ++ (for (d <- disjuncts.elements) yield d.negate)))
          //////////////////////////////////////////////////////////////////////
          
          Right(cert)
        } else {
          Left(Conjunction.FALSE)
        }
      }
    }
  }

  private def isSatisfiableGoal(tree : ProofTree) : Boolean = tree match {
    case goal : Goal => {
      val c = goal.closingConstraint
      if (c.isTrue) {
        true
      } else if (c.isFalse) {
        false
      } else {
        assert(false) // not supposed to happen
        false
      }
    }
    case _ => false
  }
  
  //////////////////////////////////////////////////////////////////////////////
  //
  // Model search in a fair manner, which means that the search in not purely
  // depth-first, but after a certain amount of time the prover switches to a
  // different branch. This is currently not used, but might sometimes by more
  // robust than the pure depth-first search
  
  private case class GoalSet(goal : Goal,
                             witnesses : List[(Substitution, TermOrder) => Substitution],
                             constsToIgnore : Set[ConstantTerm])
  
  private def findModelFair(initialGoal : Goal, stepsPerGoal : Int) : Conjunction = {
    val goals = new scala.collection.mutable.Queue[GoalSet]
    goals += GoalSet(initialGoal, List(), Set())

    while (!goals.isEmpty) {
      Timeout.check

      var GoalSet(goal, witnesses, constsToIgnore) = goals.dequeue
      
      val localGoals = new scala.collection.mutable.Stack[GoalSet]

      def destructTree(tree : ProofTree,
                       witnesses : List[(Substitution, TermOrder) => Substitution],
                       constsToIgnore : Set[ConstantTerm]) : Unit = tree match {
        case goal : Goal =>
          localGoals += GoalSet(goal, witnesses, constsToIgnore)

        case tree : WitnessTree =>
          destructTree(tree.subtree, tree.witness :: witnesses, constsToIgnore)

        case tree : QuantifiedTree => {
          //////////////////////////////////////////////////////////////////////
          Debug.assertInt(ModelSearchProver.AC, tree.quan == Quantifier.ALL)
          //////////////////////////////////////////////////////////////////////
          destructTree(tree.subtree,
                       witnesses,
                       constsToIgnore ++ tree.quantifiedConstants)
        }

        case _ : WeakenTree =>
          assert(false)  // should not happen

        case tree : ProofTreeOneChild => 
          destructTree(tree.subtree, witnesses, constsToIgnore)
     
        case tree : AndTree => {
          destructTree(tree.right, witnesses, constsToIgnore)
          destructTree(tree.left, witnesses, constsToIgnore)
        }
      }

      var i : Int = 0
      while (goal != null && i < stepsPerGoal) {
        if (isSatisfiableGoal(goal)) {
          // nothing to do here ...
          goal = null
        } else if (goal.stepPossible) {
          goal.step(ptf) match {
            case newGoal : Goal =>
              goal = newGoal
            case tree : ProofTree => {
              destructTree(tree, witnesses, constsToIgnore)
              goal = null
            }
          }
        } else {
          return constructModel(goal.facts,
                                witnesses, constsToIgnore,
                                goal.order)
        }
        
        if (goal == null && !localGoals.isEmpty) {
          val GoalSet(newGoal, newWit, newConsts) = localGoals.pop
          goal = newGoal
          witnesses = newWit
          constsToIgnore = newConsts
        }
        
        i = i + 1
      }
      
      if (goal != null)
        goals += GoalSet(goal, witnesses, constsToIgnore)
      goals ++= localGoals
    }
    
    null
  }
  
  //////////////////////////////////////////////////////////////////////////////
  
  /**
   * Construct either a countermodel or a proof for a conjecture. In case no
   * the current settings are to not produce proofs, only
   * <code>Right(List())</code> is returned for the proof case.
   */
  private def findModel(tree : ProofTree,
                        // functions to reconstruct witnesses for eliminated
                        // constants
                        witnesses : List[(Substitution, TermOrder) => Substitution],
                        // explicitly quantified constants that do not have to
                        // be included in models
                        constsToIgnore : Set[ConstantTerm],
                        depth : Int,
                        settings : GoalSettings) : Either[Conjunction, Seq[Certificate]] = {
    Timeout.check
    
    tree match {
      case goal : Goal =>
        if (isSatisfiableGoal(goal)) {
          Right(if (Param.PROOF_CONSTRUCTION(settings))
                  List(goal.getCertificate)
                else
                  List())
        } else if (goal.stepPossible)
          findModel(goal.step(ptf), witnesses, constsToIgnore, depth, settings)
        else
          Left(constructModel(goal.facts,
                              witnesses, constsToIgnore,
                              goal.order))
        
      case tree : WitnessTree =>
        findModel(tree.subtree, tree.witness :: witnesses, constsToIgnore,
                  depth, settings)

      case tree : ProofTreeOneChild => {
        ////////////////////////////////////////////////////////////////////////
        Debug.assertInt(ModelSearchProver.AC, tree match {
                          case _ : WeakenTree => false
                          case tree : QuantifiedTree => tree.quan == Quantifier.ALL
                          case _ => true
                        })
        ////////////////////////////////////////////////////////////////////////
        val newConstsToIgnore = tree match {
          case tree : QuantifiedTree => constsToIgnore ++ tree.quantifiedConstants
          case _ => constsToIgnore
        }
        findModel(tree.subtree, witnesses, newConstsToIgnore, depth, settings)
      }
     
      case AndTree(left, right, partialCert) => {
        findModel(left, witnesses, constsToIgnore, depth + 1, settings) match {
          case res @ Left(_) =>
            res
          case Right(leftCerts) =>
            findModel(right, witnesses, constsToIgnore, depth + 1, settings) match {
              case res @ Left(_) =>
                res
              case Right(rightCerts) =>
                Right(if (leftCerts.isEmpty || rightCerts.isEmpty)
                        List()
                      else if (partialCert == null)
                        leftCerts ++ rightCerts
                      else
                        List(partialCert(leftCerts ++ rightCerts)))
            }
        }
      }
    }
  }

  private def constructModel(facts : Conjunction,
                             witnesses : List[(Substitution, TermOrder) => Substitution],
                             constsToIgnore : Set[ConstantTerm],
                             order : TermOrder) : Conjunction = {
    val constantValues : Substitution =
      (new IdentitySubst(order).asInstanceOf[Substitution] /: witnesses)(
                                                         (s, w) => w(s, order))
    
    val valuesConj =
      EquationConj(for (c <- (order.orderedConstants -- constsToIgnore).elements)
                   yield LinearCombination(Array((IdealInt.ONE, c),
                                                 (IdealInt.MINUS_ONE, constantValues(c))),
                                           order),
                   order)
    Conjunction.conj(Array(valuesConj, facts), order)
  }
}

 
private case class WitnessTree(val subtree : ProofTree,
                               val eliminatedConstant : ConstantTerm,
                               val witness : (Substitution, TermOrder) => Substitution,
                               val vocabulary : Vocabulary)
                   extends { protected val subtreeOrder = vocabulary.order }
                           with ProofTreeOneChild {

  def update(newSubtree : ProofTree, newConstantFreedom : ConstantFreedom) : ProofTree =
    new WitnessTree(subtree, eliminatedConstant, witness,
                    vocabulary updateConstantFreedom newConstantFreedom)

  lazy val closingConstraint : Conjunction =
    subtree.closingConstraint

  lazy val closingConstantFreedom : ConstantFreedom =
    subtree.closingConstantFreedom

  lazy val fixedConstantFreedom : Boolean =
    subtree.fixedConstantFreedom

  lazy val stepMeaningful : Boolean =
    subtree.stepMeaningful
  
  def newConstantFreedomForSubtree(cf : ConstantFreedom) : ConstantFreedom = cf
}